
cMap = CreateClass()

local max = math.max
local min = math.min
local abs = math.abs

function GetIsoOrder (x,y) return x - y end -- = py

function MoveIsoCoordsToScreenRight 		(x,y,add) return x+add,y+add end -- =   order=x-y should stay constant



function cMap:init(px,py,w,h)
	self.x = px
	self.y = py
	self.w = w
	self.h = h
	
	self.typeStats = {}
	
	self.isoBaseX = w / 2
	self.isoBaseY = h / 2
	
	self.tileW, self.tileH = getGfxSourceSize(gImg_BaseTile)
	
	self.minIsoY = nil
	self.maxIsoY = nil
	
	self.scale = 1
	self.scaleTarget = self.scale
	self.bNoScaleSetYet = true
	
	self.tiles = {}
	self.orderrow_min = 0
	self.orderrow_max = 0
	self.orderrow = {}
	self.dynamic_objects = {} -- monster, particlesys(fire,explo,sprites), bullets, tanks...
end

function cMap:getTypeStats(tile_type)
	if not self.typeStats[tile_type] then self.typeStats[tile_type] = 0 end
	return self.typeStats[tile_type]
end

function cMap:changeTypeStats(tile_type, change)
	if not self.typeStats[tile_type] then self.typeStats[tile_type] = 0 end
	self.typeStats[tile_type] = self.typeStats[tile_type] + change
end

function cMap:setPositionOnScreen(x,y)
	self.x = x
	self.y = y
end

function cMap:getPositionOnScreen()
	return self.x, self.y
end

function cMap:addMonster(monster)
	self.dynamic_objects[monster] = true
	self.monster = monster
end

function cMap:RemoveDynmicObj	(obj) self.dynamic_objects[obj] = nil end
function cMap:AddDynmicObj		(obj) self.dynamic_objects[obj] = true end

function cMap:AddDynmicGfx(x,y,gfx,timeout)
	if (not gfx) then return end
	self.dynamic_objects[{x=x,y=y,gfx=gfx}] = true
end

function cMap:step()
	self:updateScale()
end

function cMap:setScaleByTileWidth(w)
	self:setScale(w / self.tileW)
end

function cMap:setScale(scale)
	self.scaleTarget = scale
end

function cMap:getGroundImageIndex(x,y,gGroundImageCount)
	local ax = math.abs(x)
	local ay = math.abs(y)
	
	local a = math.mod(ay*ay + ax + ay, 7) + 1
	local b = math.mod(ax*ax + ay + ax, 5) + 1
	
	if x*y < 0 then
		return math.mod(math.abs(a+b+x), gGroundImageCount) + 1
	else
		return math.mod(math.abs(a-b+y), gGroundImageCount) + 1
	end
	
end

function cMap:drawGroundSingle(isox,isoy,bx,by,sx,sy,tw,th,ox,oy)
	local px,py = self:isoToScreen(isox-ox,isoy-oy)
	local index = self:getGroundImageIndex(isox,isoy,gGroundImageCount)
	local gfx = gImg_Grounds[index]
	
	local gfxw,gfxh = getGfxSourceSize(gfx)

	-- move to big images up,  origin is left top
	if gfxh and gfxh > th then py = py - (gfxh - th) * sy end
	
	love.graphics.draw(gfx, px + bx, py + by, 0, sx, sy)
end
function cMap:drawGround(radius,bx,by,sx,sy,tw,th,ox,oy)
	for y = -radius,radius do
	for x = -radius,radius do
		local dist = math.abs(x) + math.abs(y)
		if dist <= radius then
			local isox = x+floor(ox)
			local isoy = y+floor(oy)
			self:drawGroundSingle(isox,isoy,bx,by,sx,sy,tw,th,ox,oy)
		end
	end
	end
end

function cMap:GetOrderRow(x,y)
	local order = GetIsoOrder(x,y)
	local orderrow = self.orderrow[order]
	if (not orderrow) then
		orderrow = {}
		self.orderrow[order] = orderrow
		self.orderrow_min = min(self.orderrow_min,order)
		self.orderrow_max = max(self.orderrow_max,order)
	end
	return orderrow
end

function cMap:setTile(x,y,data)
	if not self.tiles[x] then self.tiles[x] = {} end
	
	local olddata = self.tiles[x][y]
	self.tiles[x][y] = data
	
	self.minIsoY = math.min(y, self.minIsoY or y)
	self.maxIsoY = math.max(y, self.maxIsoY or y)
	
	data.isox = x
	data.isoy = y
	local orderrow = self:GetOrderRow(x,y)
	if (olddata) then orderrow[olddata] = nil end
	orderrow[data] = true
end

function cMap:getTile(x,y)
	if not self.tiles[x] then self.tiles[x] = {} end
	return self.tiles[x][y]
end

function cMap:isoToScreen(isox,isoy)
	local w = self.tileW
	local h = self.tileH
	local s = self.scale
	
	local isox_x, isox_y = w / 2, h / 2
	local isoy_x, isoy_y = w / 2, h / 2 * -1

	local px = isox * isox_x + isoy * isoy_x
	local py = isox * isox_y + isoy * isoy_y 
	
	return self.isoBaseX + px * s, self.isoBaseY + py * s -- TODO : optimize me ? function with upvalues on change to avoid field access ?
end

function cMap:updateScale()
	local d = self.maxIsoY - self.minIsoY + 1

	self:setScaleByTileWidth(self.w / d)
	
	local fps = love.timer.getFPS()
	
	self.scaleTarget = max(0.4,min(1,self.scaleTarget))
	if (self.monster) then
		self.scaleTarget = 1
	end
	
	if (self.bNoScaleSetYet) then
		self.bNoScaleSetYet = false
		self.scale = self.scaleTarget
		return
	end
	
	if self.scaleTarget ~= self.scale and fps > 0 then
		if math.abs(self.scaleTarget - self.scale) < 0.01 then
			self.scale = self.scaleTarget
		else
			local f = max(0,min(1,0.2 * 30 / fps))
			self.scale = self.scaleTarget * f + self.scale * (1 - f)
		end
	end
end

function cMap:Iso2ScreenPos(isox,isoy)
	local ox = 0
	local oy = 0
	if (self.monster) then
		ox = self.monster.x
		oy = self.monster.y
	end 
	local bx = self.x
	local by = self.y
	local px,py = self:isoToScreen(isox-ox,isoy-oy)
	return px + bx, py + by
end

function cMap:DrawImage(isox,isoy,gfx,bx,by,sx,sy,tw,th,ox,oy)
	if (not gfx) then return end
	local px,py = self:isoToScreen(isox-ox,isoy-oy)
	local gfxw,gfxh = getGfxSourceSize(gfx)
	
	-- move to big images up,  origin is left top
	if gfxh and gfxh > th then py = py - (gfxh - th) * sy end
	
	if (gfx.MyDraw) then
		gfx:MyDraw(px + bx, py + by, 0, sx, sy)
	else
		if (GetGfxZeroOrigin(gfx)) then 
			love.graphics.draw(gfx, 0,0, 0, sx, sy)
		else
			love.graphics.draw(gfx, px + bx, py + by, 0, sx, sy)
		end
	end
end

function cMap:draw()
	local tw = self.tileW
	local th = self.tileH
	
	local sx = self.scale
	local sy = self.scale
	
	
	local bx = self.x
	local by = self.y
	
	local ox = 0
	local oy = 0
	
	if (self.monster) then
		ox = self.monster.x
		oy = self.monster.y
	end
	
	
	love.graphics.setScissor(bx, by, self.w, self.h)
	love.graphics.setColor(255,255,255)
	love.graphics.setBackgroundColor(0,0,0)
	
	--~ local groundradius = math.floor(((self.maxIsoY - self.minIsoY) / 2 + 20))
	--~ self.scale = 1
	local groundradius = 10
	if (self.scale >= 0.4) then groundradius = 15 end
	if (self.scale >= 0.6) then groundradius = 12 end
	if (self.scale >= 0.8) then groundradius = 9 end
	if (self.scale >= 1.0) then groundradius = 8 end
	--~ print("groundradius",self.scale,groundradius)
	--~ self:drawGround(groundradius,bx,by,sx,sy,tw,th,ox,oy)
	
	local orderrow_min = self.orderrow_min
	local orderrow_max = self.orderrow_max
	for obj,v in pairs(self.dynamic_objects) do
		local order = GetIsoOrder(floor(obj.x+0.5),floor(obj.y+0.5))
		obj.order = order
		orderrow_min = min(orderrow_min,order)
		orderrow_max = max(orderrow_max,order)
		assert(obj.gfx)
	end
	
	local gox,goy = floor(ox),floor(oy)
	local gorder = GetIsoOrder(gox,goy)
	orderrow_min = min(orderrow_min,gorder-groundradius)
	orderrow_max = max(orderrow_max,gorder+groundradius)
	
	for order = orderrow_min,orderrow_max do 
		-- draw ground
		local gx,gy = gox+(order-gorder+1),goy
		for add=-groundradius,groundradius do 
			local x,y = MoveIsoCoordsToScreenRight(gx,gy,add)
			self:drawGroundSingle(x,y,bx,by,sx,sy,tw,th,ox,oy)
		end
		
		-- draw buildings
		for tile,v in pairs(self.orderrow[order] or {}) do
			self:DrawImage(tile.isox,tile.isoy,tile.img or gImg_BaseTile,bx,by,sx,sy,tw,th,ox,oy)
		end
		
		-- DrawDynamicObjects
		for obj,v in pairs(self.dynamic_objects) do 
			if (obj.order == order and (not obj.bDrawFinal)) then 
				-- todo : image anim ? 
				self:DrawImage(obj.x,obj.y,obj.gfx,bx,by,sx,sy,tw,th,ox,oy)
			end
		end
	end
	
	
	for obj,v in pairs(self.dynamic_objects) do 
		if (obj.bDrawFinal) then 
			self:DrawImage(obj.x,obj.y,obj.gfx,bx,by,sx,sy,tw,th,ox,oy)
		end
	end
	
	completeScreenOnly()
end


