
local max = math.max
local min = math.min
local abs = math.abs


gTurnSpeed_Monster = 1.1*math.pi
gSpeed_Monster = 1.1
gSpeed_FireBall = 5.2

gMonster_ChargeSpeedFactor = 4

gMonster_AttackDelay = 1000
gMonster_AttackAnimLen = 500
gMonster_ChargeLen = 1000

cMonster = CreateClass()

function cMonster:init(map, res, ind, com)
	self.map = map
	self.x=0
	self.y=0
	self.ang=0
	self.gfx=gImg_Monster
	self.attack_charge_until = 0
	--~ print("MONSTER CREATE",res,ind,com)
	self.score_res = res -- 0 bis 100
	self.score_ind = ind
	self.score_com = com
	
	self.next_special = 0
	self.next_walk = 0
	self.name = "???????"
	
	self.points_casualties = 0
	self.points_damages = 0
end

function cMonster:destroyBuilding(x,y)
	local tile = self.map:getTile(x,y)
	if tile and not tile.bDestroyed then
		self:addPointByType(tile.type)
	end
	
	self.map:DestroyBuilding(x,y)
end

function cMonster:MonsterStep(keymap)
	self.name = (self.map == gMap1) and "monstA" or "monstB"
	
	local add =		(love.keyboard.isDown(keymap[1]) and -1 or 0) + 
					(love.keyboard.isDown(keymap[2]) and  1 or 0) 
			
	local bWalk = gMyTicks > self.next_walk
	local bCharge = gMyTicks < self.attack_charge_until
	
					
	local monster = self
	monster.ang = monster.ang + gDeltaTime*add*gTurnSpeed_Monster
	local v = gDeltaTime * gSpeed_Monster
	
	local walk_anim_speed_factor = 1
	if (bCharge) then 
		walk_anim_speed_factor = gMonster_ChargeSpeedFactor
		v = v * gMonster_ChargeSpeedFactor 
		
		-- smoke and fire behind charge
		if (math.random() < 0.4) then 
			local r = 0.2
			self.map:AddDynmicGfx(self.x+frand(-r,r),self.y+frand(-r,r),(math.random() < 0.1) and MakePS_SmokeShort() or MakePS_BurnShort())
		end
	end
	
	local vx = math.sin(monster.ang)
	local vy = math.cos(monster.ang)
	if (bWalk) then 
		monster.x = monster.x + v*vx
		monster.y = monster.y + v*vy
	end
	local dir = (abs(vx) > abs(vy)) and (vx > 0 and "e" or "w") or (vy > 0 and "n" or "s")
	
	
	local imgarr = gImgMonster1[dir]
	
	local imgindex = 1 + math.mod(math.floor(gMyTicks/500*walk_anim_speed_factor),4)
	if (not bWalk) then imgindex = self.nonwalk_img_index end -- attack
	monster.gfx = imgarr[imgindex]
	
	
	local rounded_x = floor(monster.x)
	local rounded_y = floor(monster.y)
	if (monster.rounded_x ~= rounded_x or 
		monster.rounded_y ~= rounded_y) then
		monster.rounded_x  = rounded_x
		monster.rounded_y  = rounded_y
		-- monster walks into new tile
		self:destroyBuilding(rounded_x  ,rounded_y  )
		self:destroyBuilding(rounded_x+1,rounded_y  )
		self:destroyBuilding(rounded_x  ,rounded_y+1)
		self:destroyBuilding(rounded_x+1,rounded_y+1)
	end
	
	if (love.keyboard.isDown(keymap[3]) and gMyTicks > self.next_special) then
		self.next_special = gMyTicks + gMonster_AttackDelay
		
		--~ print("monster attack",self.name,self.score_res,self.score_ind,self.score_com)
		random_do({
			[function () self:Attack_FireBall(vx,vy) end]=1+self.score_res, -- key=fun,val=probability  score_x:0 bis 100
			[function () self:Attack_Charge(vx,vy) end]=1+self.score_ind,
			[function () self:Attack_Smash(vx,vy) end]=1+self.score_com,
		})
	end
end


function cMonster:Attack_FireBall(vx,vy)
	PlaySound(gSound_monster_scream)
	self.nonwalk_img_index = 6
	local speed = gSpeed_FireBall
	self.next_walk = gMyTicks + gMonster_AttackAnimLen
	CreateFireBall(self,self.x,self.y,speed*vx,speed*vy,500,0)
end
function cMonster:Attack_Smash(vx,vy)
	PlaySound(gSound_monster_scream_2)
	self.nonwalk_img_index = 5
	self.next_walk = gMyTicks + gMonster_AttackAnimLen + 500
	InvokeLater(function () self:DestroyBuildingsInRadius(self.x,self.y,2) end,500)
	self.map:AddDynmicGfx(self.x,self.y,MakePS_Dustcircle()) 
end
function cMonster:Attack_Charge(vx,vy)
	PlaySound(gSound_monster_scream_3)
	--~ self.nonwalk_img_index = 6
	self.attack_charge_until = gMyTicks + gMonster_ChargeLen
	
	
	--~ local speed = gSpeed_FireBall 
	--~ CreateFireBall(self,self.x,self.y,speed*vx,speed*vy,1300,0)
end

function InvokeLater (fun,delay) 
	local endt = gMyTicks + delay
	RegisterStepMe({Step=function() if (gMyTicks > endt) then fun() return true end end})
end

function frand (a,b) return a+(b-a)*math.random() end
function CreateFireBall (monster,x,y,vx,vy,lifetime)
	--~ print("CreateFireBall",x,y,vx,vy,lifetime)
	local o = {
		bDrawFinal=true,
		map=monster.map,
		monster=monster,
		x=0,
		y=0,
		px=x,
		py=y,
		vx=vx,
		vy=vy,
		--~ gfx=gImgFireBall,
		gfx=MakePS_FireBall(),
		endt=gMyTicks+lifetime,
		Step=function (self) 
			self.px = self.px + self.vx * gDeltaTime
			self.py = self.py + self.vy * gDeltaTime
			--~ self.x = self.px
			--~ self.y = self.py
			local x,y = self.map:Iso2ScreenPos(self.px,self.py)
			self.gfx:setPosition(floor(x),floor(y))
			--~ print("FireBallStep",self.x,self.y)
			if (gMyTicks > self.endt) then
				--~ print("FireBall END")
				InvokeLater(function () self.monster:DestroyBuildingsInRadius(self.px,self.py,1) end,300)
				self.map:RemoveDynmicObj(self)
				
				local r = 1
				self.map:AddDynmicGfx(self.px,self.py,MakePS_Firecircle()) 
				for i=1,3 do 
					InvokeLater(function ()
						for i=1,5 do self.map:AddDynmicGfx(self.px+frand(-r,r),self.py+frand(-r,r),MakePS_BurnShort()) end
						for i=1,1 do self.map:AddDynmicGfx(self.px+frand(-r,r),self.py+frand(-r,r),MakePS_SmokeShort()) end
					end,i*200)
				end
				
				return true
			end
		end,
	}
	monster.map:AddDynmicObj(o)
	RegisterStepMe(o)
end

function cMonster:DestroyBuildingsInRadius(x,y,r)
	r = math.ceil(r)
	x = math.floor(x)
	y = math.floor(y)
	for rx=-r,r do
	for ry=-r,r do
	self:destroyBuilding(x+rx,y+ry)
	end
	end
end
function cMap:DestroyBuildingsInRadius(x,y,r)
	r = math.ceil(r)
	x = math.floor(x)
	y = math.floor(y)
	for rx=-r,r do
	for ry=-r,r do
	self:DestroyBuilding(x+rx,y+ry)
	end
	end
end

function cMonster:addPointByType(tile_type)
	if tile_type == 1 then
		self.points_casualties = self.points_casualties + 50
		self.points_damages = self.points_damages + 0
	elseif tile_type == 2 then
		self.points_casualties = self.points_casualties + 0
		self.points_damages = self.points_damages + 50
	elseif tile_type == 3 then
		self.points_casualties = self.points_casualties + 25
		self.points_damages = self.points_damages + 25
	end
	
end

function cMap:DestroyBuilding(x,y)
	x = floor(x)
	y = floor(y)
	--~ print("cMap:DestroyBuilding",x,y)
	local tile = self:getTile(x,y)
	if (tile and (not tile.bDestroyed)) then 	
		tile.bDestroyed = true
		
		local t = tile.type
		
		-- set new graphic
		tile.img = gDestroyedImage[tile.img] or tile.img
		
		-- if the thing destroyed was a building, make some effects
		if (t == kBuildingType_Residence or 
			t == kBuildingType_Industry or 
			t == kBuildingType_Commerce ) then
		
			-- building destroyed
			
			random_do({
				[function () PlaySound(gSound_explo1) end]=1, -- key=fun,val=probability
				[function () PlaySound(gSound_explo2) end]=1,
				[function () PlaySound(gSound_explo3) end]=1,
			})
	
			random_do({
				[function () 
					--self:AddDynmicGfx(x,y,MakePS_Smoke()) 
					self:AddDynmicGfx(x,y,MakePS_Burn()) 
					StartGlobalFireSound()
				end]=2,
				[function () 
					self:AddDynmicGfx(x,y,MakePS_Smoke()) 
					self:AddDynmicGfx(x,y,MakePS_Burn()) 
					StartGlobalFireSound()
				end]=1,
				[function () 
				
				end]=3,
			})

			
			if math.random () <= 0.03 then 
				random_do({
					[function () PlaySound(gSound_scream1) end]=1, -- key=fun,val=probability
					[function () PlaySound(gSound_scream2) end]=1,
					[function () PlaySound(gSound_scream3) end]=1,
				})
			end
			
		end
		
	end
end

function StartGlobalFireSound()
	if (gNextGlobalFireSoundStep) then return end
	gNextGlobalFireSoundStep = 0
	-- the following is only done once
	
end

function StepGlobalFireSound()
	if (not gNextGlobalFireSoundStep) then return end
	if (gNextGlobalFireSoundStep > gMyTicks) then return end
	random_do({
		[function () PlaySound(gSound_fire1) gNextGlobalFireSoundStep = gMyTicks + 4000 end]=1, -- key=fun,val=probability
		[function () PlaySound(gSound_fire2) gNextGlobalFireSoundStep = gMyTicks + 5000 end]=1,
	})
end
